/*
 * sig.c
 *
 * The Quoteriser's random signature generator.
 *
 *      Created: 8th December 1997
 * Version 1.00: 21st December 1997
 * Version 1.01: 6th June 1998
 *
 * (C) 1997 Nicholas Paul Sheppard
 *
 * This file is distributed under the GNU General Public License. See the
 * file copying.txt for details.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <io.h>
#include <ctype.h>
#include <time.h>
#include "quotes.h"
#include "authors.h"
#include "html.h"
#include "general.h"
#include "typeset.h"


/* limits */
#define SIG_MAX_CELLS	80

/* internal function prototypes */
void	SigGenerate(QUOTEDB *, AUTHORDB *, char *, int, FILE *);
char	*SigGenerateRow(QUOTEDB *, AUTHORDB *, char *, int, FILE *);
int	SigQuoteTag(char *);
char	*SigRandomQuote(QUOTEDB *, AUTHORDB *, int, int);


int main(int argc, char **argv)
{
	FILE		*fTemplate, *fSig;
	char		*pszQuoteDB, *pszAuthorDB, *pszTemplate;
	QUOTEDB		qdb;
	AUTHORDB	adb;
	int		bPrintAuthor, bPrintSource, i;

	/* set defaults */
	bPrintAuthor = 1;
	bPrintSource = 1;

	/* parse command line */
	i = 1;
	while ((i < argc) && ((argv[i][0] == '-') || (argv[i][0] == '/'))) {
		if (argv[i][1] == 'a') {
			bPrintAuthor = 0;
		} else if (argv[i][1] == 's') {
			bPrintSource = 0;
		} else if (argv[i][1] == '?') {
			i = argc;
		} else if (argv[i][1] == '-') {
			i++;
			break;
		} else {
			fprintf(stderr, "Unrecognised option %s.\n", argv[i]);
		}
		i++;
	}


	if (i >= argc) {
		/* print help screen */
		printf("The Quoteriser: Random Signature Generator 1.00\n");
		printf("(C) 1997 Nicholas Paul Sheppard. This program is distributed\n");
		printf("under the GNU General Public License and comes with NO WARRANTY\n");
		printf("WHATSOEVER. See the file copying.txt for details.\n\n");
		printf("Usage: sig [options] <template> [<signature>]\n\n");
		printf("  where: <template>  is the name of the template (input) file\n");
		printf("         <signature> is the name of the signature (output) file\n\n");
		printf("Valid options:\n");
		printf("  -?  print this message\n");
		printf("  -a  don't print author\n");
		printf("  -s  don't print quote source\n");
		printf("  --  no more arguments on the command line\n\n");
		printf("Environment: QUOTER_QDB - quote database (must be set)\n");
		printf("             QUOTER_ADB - author database (optional)\n");
		return (0);
	}

	/* open quote database */
	if ((pszQuoteDB = getenv("QUOTER_QDB")) == NULL) {
		fprintf(stderr, "The QUOTER_QDB environment variable is not set.\n");
		return (1);
	}
	if (!QuoteOpenDB(pszQuoteDB, &qdb, S_IREAD)) {
		fprintf(stderr, "sig: unable to open %s for reading.\n", pszQuoteDB);
		return (1);
	}

	/* open author database (if necessary) */
	if (!bPrintAuthor || ((pszAuthorDB = getenv("QUOTER_ADB")) == NULL)) {
		AuthorNullifyDB(&adb);
	} else if (!AuthorOpenDB(pszAuthorDB, &adb, S_IREAD)) {
		fprintf(stderr, "sig: unable to open %s for reading.\n", pszAuthorDB);
		return (1);
	}

	/* open template file */
	if ((fTemplate = fopen(argv[i], "r")) == NULL) {
		fprintf(stderr, "sig: %s: %s", argv[i], strerror(errno));
		return (1);
	}

	/* open signature file */
	if ((i + 1) >= argc) {
		/* no signature file specified; use stdout */
		fSig = stdout;
	} else {
		if ((fSig = fopen(argv[i + 1], "w")) == NULL) {
			fprintf(stderr, "sig: %s: %s", argv[i + 1], strerror(errno));
			return (1);
		}
	}

	/* read template into memory */
	if ((pszTemplate = strfromf(fTemplate)) == NULL) {
		fprintf(stderr, "sig: memory allocation failure.\n");
		return (1);
	}

	/* do it */
	srand(time(NULL));
	SigGenerate(&qdb, &adb, pszTemplate, bPrintSource, fSig);

	/* close files */
	QuoteCloseDB(&qdb);
	if (adb.dbfInfo != NULL)
		AuthorCloseDB(&adb);
	fclose(fTemplate);
	if (fSig != stdout)
		fclose(fSig);

	/* free memory */
	free(pszTemplate);

	return (0);
}


void SigGenerate(QUOTEDB *pqdb, AUTHORDB *padb, char *pszTemplate, int bPrintSource, FILE *f)
/*
 * Generate a signature file, inserting random quotes where indicated by the
 * template string.
 *
 * QUOTEDB *pqdb	- quote database to obtain quotes from
 * AUTHORDB *padb	- author database to contain authors from (may be empty)
 * char *pszTemplate	- template string - will be altered
 * int bPrintSource	- print the source of the quote?
 * FILE *f		- output file
 */
{
	char	*pszCurrent, *pszNext, szTag[HTML_MAX_TAG];
	int	i;

	pszCurrent = pszTemplate;
	while (pszCurrent[0] != '\0') {

		/* search for next tag */
		while ((pszCurrent[0] != '\0') && (pszCurrent[0] != '<'))
			pszCurrent++;

		/* hopefully, it's a <TR> tag */
		if (pszCurrent[0] == '<') {
			i = HTMLGetNextChunk(pszCurrent, szTag, &pszNext);
			pszCurrent = pszNext;
			if ((i == HTML_TAG_MID) || (i == HTML_TAG_END)) {
				if (HTMLParseTag(szTag) == HTML_TABLE_ROW) {
					/* format the row */
					pszNext = SigGenerateRow(pqdb, padb, pszCurrent, bPrintSource, f);
					pszCurrent = pszNext;
				}
			}
		}
	}
}


char *SigGenerateRow(QUOTEDB *pqdb, AUTHORDB *padb, char *pszTemplate, int bPrintSource, FILE *f)
/*
 * Typeset one row of cells.
 *
 * QUOTEDB *pqdb	- quote database to obtain quotes from
 * AUTHORDB *padb	- author database to contain authors from (may be empty)
 * char *pszTemplate	- template string - will be altered
 * int bPrintSource	- print the source of the quote?
 * FILE *f		- output file
 *
 * Returns		- pointer to the position in pszTemplate from which to carry on
 */
{
	char		*apszCells[SIG_MAX_CELLS], *pszCurrent, szTag[HTML_MAX_TAG], *pszNext;
	int		aiWidth[SIG_MAX_CELLS], iCount, i;
	HTML_TDATTR	tda;
	int		bDone, bMove, abAllocated[SIG_MAX_CELLS];

	/* divide the input string up into cells */
	pszCurrent = pszTemplate;
	iCount = 0;
	bDone = 0;
	memset(abAllocated, 0, SIG_MAX_CELLS * sizeof(int));
	do {
		/* search for next tag */
		while ((pszCurrent[0] != '\0') && (pszCurrent[0] != '<'))
			pszCurrent++;

		if (pszCurrent[0] == '<') {
			bMove = 1;
			switch (HTMLGetNextChunk(pszCurrent, szTag, &pszNext)) {
				case HTML_TAG_MID:
				case HTML_TAG_END:
					switch (HTMLParseTag(szTag)) {
						case HTML_TABLE_ROW:
							bDone = 1;
							bMove = 0;
							break;

						case HTML_TABLE_DATA:
							HTMLAttrTD(szTag, &tda);
							apszCells[iCount] = pszNext;
							aiWidth[iCount] = tda.iColSpan;
							if (++iCount > SIG_MAX_CELLS)
								bDone = 1;
							break;

						case HTML_UNKNOWN:
							/* hopefully, a <quote> tag */
							if ((aiWidth[iCount] = SigQuoteTag(szTag)) > 0) {
								apszCells[iCount] = SigRandomQuote(pqdb, padb, aiWidth[iCount], bPrintSource);
								abAllocated[iCount] = 1;
								if (++iCount > SIG_MAX_CELLS)
									bDone = 1;
							}
							break;
					}
					break;

				case HTML_END:
					bDone = 1;
					break;
			}
			*(pszCurrent - 1) = '\0';
			if (bMove)
				pszCurrent = pszNext;
		} else
			bDone = 1;
	} while (!bDone);

	/* write the row cells out to the file */
	if (iCount > 0)
		if (!strboxf(f, apszCells, aiWidth, iCount)) {
			fprintf(stderr, "sig: memory allocation failure.\n");
			exit(1);
		}

	/* de-allocate any memory we allocated */
	for (i = 0; i < iCount; i++)
		if (abAllocated[i])
			free(apszCells[i]);

	return (pszCurrent);
}


int SigQuoteTag(char *pszTag)
/*
 * Read information from a <quote> tag. This is *not* a real HTML tag;
 * this is used for Quoteriser purposes only.
 *
 * char *pszTag		- the tag
 *
 * Returns		- the value of the COLSPAN attribute
 *			  0 if there is a problem
 */
{
	char *pch, szString[HTML_MAX_TAG], szAttr[8];
	int i;

	/* find tag identifier */
	pch = pszTag + 1;
	while (isspace(*pch))
		pch++;

	/* check that this *is* a quote tag */
	i = 0;
	while (!isspace(*pch) && ((*pch) != '>')) {
		szString[i++] = (*pch);
		pch++;
	}
	szString[i] = '\0';
	if (stricmp(szString, "QUOTE") != 0)
		/* not a <quote> tag */
		return (0);

	/* look for COLSPAN attribute */
	i = 0;
	pch = HTMLAttrFirst(pszTag, szAttr, sizeof(szAttr), szString, sizeof(szString));
	while (pch != NULL) {
		if (stricmp(szAttr, "COLSPAN") == 0)
			i = atoi(szString);
		pch = HTMLAttrNext(pch, szAttr, sizeof(szAttr), szString, sizeof(szString));
	}

	return (i);
}


char *SigRandomQuote(QUOTEDB *pqdb, AUTHORDB *padb, int iWidth, int bPrintSource)
/*
 * Choose and typeset a random quote. Memory for the quote will be
 * allocated.
 *
 * QUOTEDB *pqdb	- quote database to get quote from
 * AUTHORDB *padb	- author database to get author from
 * int bPrintSource	- print the source of the quote?
 * int iWidth		- width to typeset quote in
 *
 * Returns	- a pointer to the quote
 *		  NULL if there is a problem
 */
{
	QUOTESEARCH	qs;
	QUOTEINFO	*pqi;
	AUTHORINFO	*pai;
	int		i, iQuote;
	int		bPrintAuthor;
	char		*pszText, *apszTypeset[3], szString[210], *pszRet;

   	/* count the number of quotes in the database */
	i = 0;
	if (QuoteFullSearchInit(pqdb, &qs)) {
		i++;
		while (QuoteFullSearchNext(&qs))
			i++;
	}
	QuoteFullSearchEnd(&qs);

	if (i == 0) {
		/* empty database */
		return (NULL);
	}

	/* choose a random quote */
	if (i > 1) {
		iQuote = rand() % i;
	} else
		iQuote = 0;
	for (QuoteFullSearchInit(pqdb, &qs), i = 0; i < iQuote; QuoteFullSearchNext(&qs), i++);
	QuoteFullSearchEnd(&qs);

	/* initialise the typeset array to be empty */
	for (i = 0; i < 3; i++)
		apszTypeset[i] = NULL;

	/* typeset the quote */
	bPrintAuthor = 0;
	if (QuoteGetQuote(pqdb, qs.szLastCode, &pqi, &pszText)) {
		if (padb->dbfInfo != NULL)
			if (AuthorGetAuthor(padb, pqi->szAuthorCode, &pai, NULL))
				bPrintAuthor = 1;
	} else
		pai = NULL;
	if (pszText != NULL)
		apszTypeset[0] = pszText;
	bPrintSource = bPrintSource && (strlen(pqi->szSource) > 0);
	if (bPrintAuthor || bPrintSource)  {
		strcpy(szString, "<p>- ");
		if (bPrintAuthor) {
			strcat(szString, pai->szGivenName);
			if ((strlen(pai->szSurname) > 0) && (strlen(pai->szGivenName) > 0))
				strcat(szString, " ");
			strcat(szString, pai->szSurname);
		}
		if (bPrintAuthor && bPrintSource)
			strcat(szString, ", ");
		if (bPrintSource) {
			strcat(szString, "<i>");
			strcat(szString, pqi->szSource);
			strcat(szString, "</i>");
		}
		apszTypeset[1] = szString;
	}
	pszRet = TypesetASCII(apszTypeset, iWidth);

	/* de-allocate memory */
	QuoteFreeQuote(&pqi, &pszText);
	if (bPrintAuthor)
		AuthorFreeAuthor(&pai, NULL);

	return (pszRet);
}
