/*
 * str.c
 *
 * Miscellaneous string-handling functions.
 *
 *      Created: 25th November, 1997 (was part of general.c)
 * Verison 2.00: 15th December, 1997
 *
 * (C) 1997 Nicholas Paul Sheppard
 *
 * This file is distributed under the GNU General Public License. See the
 * file copying.txt for details.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "general.h"

char *strabbrev(char *pszString, char *pszAbbrev, int iMaxLength)
/*
 * Truncate a string to iMaxLength characters, breaking at word boundaries only,
 * and appending an ellipsis to the truncated string if necessary.
 *
 * char *pszString	- the string to be abbreviated
 * char *pszAbbrev	- the abbreviated version of the string (output)
 * int iMaxLength	- the maximum length to abbreviate to (including any ellipses)
 *
 * Returns:		- pszAbbrev
 */
{
	char *pchSave, *pchProbe;
	int bDone, bEllipsis;

	/* start at the beginning of the string */
	pchProbe = pchSave = pszString;

	/* go through the string word-by-word until we go as far as we can */
	bDone = 0;
	bEllipsis = 1;
	while (!bDone) {

		/* find the end of the next word */
		while (!isspace(*pchProbe) && !ispunct(*pchProbe) && ((*pchProbe) != '\0'))
			pchProbe++;

		if (((*pchProbe) == '\0') && ((pchProbe - pszString) < iMaxLength)) {
			/* pszString has ended; stop and don't use an ellipsis */
			pchSave = pchProbe;
			bDone = 1;
			bEllipsis = 0;
		} if ((pchProbe - pszString) < (iMaxLength - 3)) {
			/* still inside limit; continue */
			pchSave = pchProbe++;
		} else {
			/* gone past the limit; stop and use an ellipsis */
			bDone = 1;
			bEllipsis = 1;
		}
	}

	/* form abbreviated string */
	strncpy(pszAbbrev, pszString, pchSave - pszString);
	pszAbbrev[pchSave - pszString] = '\0';
	if (bEllipsis)
		strcat(pszAbbrev, "...");

	return (pszAbbrev);
}


int strboxf(FILE *f, char **papszBox, int *aiWidth, int n)
/*
 * Print a series of strings to a file in the form of a series of boxes.
 * The first string in papsz will be in the leftmost box, and the last (nth)
 * string in the right-most box. Strings in excess of the given width for
 * their box will be truncated to fit.
 *
 * FILE *f		- the file to print to
 * char **papszBox	- the array of strings to put in the boxes
 * int *aiWidth		- array of widths for each box
 * int n		- the number of boxes
 *
 * Returns		- 1 if success
 *                        0 if failure (memory allocation)
 */
{
	char *pszLine, **apchCurrent, *pch;
	int i, iCount;
	int bDone;

	/* stop now if papszBox is trivial */
	if (n < 1)
		return (1);

	/* allocate memory to hold a line (plus two bytes for EOL stuff) */
	iCount = 2;
	for (i = 0; i < n; i++)
		iCount += aiWidth[i];
	if ((pszLine = (char *)malloc(iCount)) == NULL)
		return (0);

	/* initialise cursor array */
	if ((apchCurrent = (char **)malloc(n)) == NULL) {
		free(pszLine);
		return (0);
	}
	for (i = 0; i < n; i++)
		apchCurrent[i] = papszBox[i];

	do {
		/* build up a line */
		pszLine[0] = '\0';
		bDone = 1;
		for (i = 0; i < n; i++) {
			if ((*apchCurrent[i]) != '\0') {

				/* still have some to go, so set done to false */
				bDone = 0;

				/* determine what goes in this box for this line */
				if ((pch = strchr(apchCurrent[i], '\n')) == NULL)
					pch = strchr(apchCurrent[i], '\0');
				iCount = pch - apchCurrent[i];
				if (iCount > aiWidth[i])
					iCount = aiWidth[i];

				/* append it to the line we're working on */
				strncat(pszLine, apchCurrent[i], iCount);

				/* update cursor for this box */
				apchCurrent[i] = pch;
				if ((*pch) == '\n')
					apchCurrent[i]++;

				/* pad with spaces out to the width of the box */
				pch = strchr(pszLine, '\0');
				for (iCount = aiWidth[i] - iCount; iCount > 0; iCount--) {
					(*pch) = ' ';
					pch++;
				}
				(*pch) = '\0';
			} else {
				pch = strchr(pszLine, '\0');
				for (iCount = aiWidth[i]; iCount > 0; iCount--) {
					(*pch) = ' ';
					pch++;
				}
				(*pch) = '\0';
			}
		}

		if (!bDone) {
			/* add the line to the file */
			strcat(pszLine, "\n");
			fputs(pszLine, f);
		}
	} while (!bDone);

	/* clean up */
	free(pszLine);
	free(apchCurrent);

	return (1);
}


char *strencl(const char *pszString, char *pszDelimited, char chLeftDelim, char chRightDelim)
/*
 * Find text in a substring enclosed by given delimiters. The last set of delimiters will
 * be used.
 *
 * char *pszString	- the string to search in
 * char *pszDelimited	- the delimited substring (output)
 * char chLeftDelim	- the left-hand delimiter
 * char chRightDelim	- the right-hand delimiter
 *
 * Returns:		- pszDelimited if a pair of delimiters are found
 *			  NULL otherwise
 */
{
	char *pchLeft, *pchRight;

	if ((pchLeft = strrchr(pszString, chLeftDelim)) != NULL)
		if ((pchRight = strchr(pchLeft + 1, chRightDelim)) != NULL) {
			strncpy(pszDelimited, pchLeft + 1, pchRight - pchLeft - 1);
			pszDelimited[pchRight - pchLeft - 1] = '\0';
			return (pszDelimited);
		}

	return (NULL);
}


char *strfchr(char *pszString, char *pchMarker, char ch)
/*
 * See if a character appears in a string before a given position.
 *
 * char *pszString	- the string to search in
 * char *pchMarker	- the current position in the string
 * char ch		- the character to search for
 *
 * Returns:		- a pointer to ch in the string, if it occurs before pchMarker
 *			  pchMarker otherwise
 */
{
	char *pch;

	pch = strchr(pszString, ch);
	if ((pchMarker == NULL) || ((pch != NULL) && (pch < pchMarker)))
		return (pch);
	else
		return (pchMarker);
}


char *strfromf(FILE *f)
/*
 * Place the contents of a file into a null-terminated string. Space for the string will
 * be allocated.
 *
 * FILE *f		- the file to be read into the buffer
 *
 * Returns		- NULL if there is a memory allocation failure
 *			  a pointer to the allocated string otherwise
 */
{
	char *pszRet;
	char szLine[1000];
	long lPos;

	/* figure out how much space we need to allocate for the string */
	lPos = ftell(f);
	fseek(f, 0L, SEEK_END);
	if ((pszRet = (char *)malloc(ftell(f))) == NULL)
		return (NULL);
	fseek(f, lPos, SEEK_SET);

	/* compile the string */
	pszRet[0] = '\0';
	while (!feof(f)) {
		if (fgets(szLine, 1000, f) == NULL)
			break;
		strcat(pszRet, szLine);
	}

	return (pszRet);
}


char *strpre(char *pszString, char *pszInsert)
/*
 * Pre-pend one string to another.
 *
 * char *pszString	- the string to be pre-pended to
 * char *pszInsert	- the string to be pre-pended
 *
 * Returns		- pszString
 */
{
	char	*pch;
	int	n;

	/* move pszString forward to make space */
	n = strlen(pszInsert);
	for (pch = strchr(pszString, '\0'); pch >= pszString; pch--)
		*(pch + n) = *pch;

	/* copy pszInsert into the space we've created */
	strncpy(pszString, pszInsert, n);

	return (pszString);
}


char *strreplace(char *pszTemplate, const char *pszMarker, const char *pszReplace, char *pszOutput)
/*
 * Replace every occurrence of a marker in a string with a value.
 *
 * char *pszTemplate	- the template string containing markers
 * char *pszMarker	- the marker to be replaced
 * char *pszReplace	- the string to replace the marker with
 * char *pszOutput	- a string to hold the resulting string (output)
 *
 * Returns		- pszOutput
 */
{
	char *pchStart, *pchEnd;

	pszOutput[0] = '\0';
	pchStart = pszTemplate;
	while ((pchEnd = strstr(pchStart, pszMarker)) != NULL) {
		strncat(pszOutput, pchStart, pchEnd - pchStart);
		strcat(pszOutput, pszReplace);
		pchStart += strlen(pszMarker);
	}
	strcat(pszOutput, pchStart);

	return (pszOutput);
}
